<?
$dir = dirname(__FILE__);
putenv("PATH=/bin:/usr/bin:/usr/local/bin:$dir");

define('NV_DEFAULT_RESOLUTION', "1280x800");
$nvidia_displays = get_displays();
$nvidia_metamodes = get_metamodes();

function get_metamodes()
{
    exec('nv-get-metamodes', $metamodes_str);

    $metamodes = array();
    foreach($metamodes_str as $str) {
        $metamode_info = parse_metamode_line($str);
        $metamodes[$metamode_info['id']] = $metamode_info;
    }

    return $metamodes;
}

function get_current_metamode()
{
    exec('nv-get-current-metamode', $id);
    return $id;

}

function get_displays()
{
    exec('nv-get-displays', $output);

    $raw_displays = array_shift($output);
    $raw_displays = explode(' ', $raw_displays);
    
    foreach($raw_displays as $display) {
        $display = explode(',', $display);
        $displays[$display[0]] = intval(substr($display[1],2), 16);
    }

    return $displays;
}

function get_default_metamode($metamodes)
{
    // search all existing metamodes to find the DEFAULT mode
    // We define the default mode as the one where only the laptop screen is on

    $default = null;

    foreach($metamodes as $metamode_info) {
        // find default "lcd only" mode
        if(
                $metamode_info['displays']['DFP-0'] &&
                $metamode_info['displays']['DFP-0']['enabled']
        ) {
            $is_target = true;

            // search other displays and make sure they are disabled
            foreach($metamode_info['displays'] as $display => $display_info) {
                if($display == "DFP-0") continue;

                // if this is the default mode, all other displays must be disabled
                if($display_info['enabled']) {
                    $is_target = false;
                    break;
                }
            }

            if($is_target) {
                $default = $metamode;
                break;
            }
        }
    }

    return $default;
}

// parse output line of nv_get_metamodes command
function parse_metamode_line($str) {

    // line is in form "<id>;<metamode>"
    list($id, $metamode) = explode(';', $str);

    $fields = array();
    $fields['metamode'] = $metamode;
    $fields['id'] = $id;
    $fields['displays'] = parse_metamode($metamode);

    return $fields;
}

// parse metamode string into an array
function parse_metamode($str) {
    $displays = explode(", ", $str);
    $parsed_displays = array();

    foreach($displays as $display) {
        $display_info = array();
        list(
            $key, // CRT-0, DFP-0, etc
            $display_info['modeline'], // NULL, or multi-parameter modeline
        ) = explode(': ', $display);

        $display_info['display'] = $key;

        if($display_info['modeline'] == "NULL") {
            $display_info['enabled'] = false;
        } else {
            $display_info['enabled'] = true;

            list(
                $display_info['mode'], // usually "nvidia-auto-select"
                $display_info['resolution'], // e.g. "@1024x768"
                $display_info['offset'] // e.g. "+0+0"
            ) = explode(' ', $display_info['modeline']);

            // strip off leading '@'
            $display_info['resolution'] = substr($display_info['resolution'],1);
        }

        $parsed_displays[$key] = $display_info;
    }

    return $parsed_displays;
}

function create_default_metamode($available_displays) 
{
    $dfp_display = array(
                'display' => 'DFP-0',
                'enabled' => true,
                'mode' => 'nvidia-auto-select',
                'resolution' => '1440x900',
                'offset' => '+0+0'
    );

    $displays = array('DFP-0' => $dfp_display);

    foreach($available_displays as $display) {
        if($display == 'DFP-0') continue;

        $displays[$display] = array(
                'display' => $display,
                'enabled' => false
        );
    }

    $metamode = create_metamode_string($displays);

    $id = add_metamode($metamode);

    $default_metamode = array(
        'metamode' => $metamode,
        'id' => $id,
        'displays' => $displays
    );


    return $default_metamode;
}


function switch_mode($id)
{
    global $nvidia_metamodes;
    $mode_info = $nvidia_metamodes[$id];

    // delete existing metamodes (there shouldn't be any, but just in case...)
    $current_metamode = get_current_metamode();

    // get total screen resolution
    $width = 0;
    $height = 0;
    foreach($mode_info['displays'] as $display) {
        if(!$display['enabled']) continue;

        $offsets = explode('+', $display['offset']);
        $resolution = explode('x', strtolower($display['resolution']));

        $width = max($width, (int) $offset[0] + (int) $resolution[0]);
        $height = max($height, (int) $offset[1] + (int) $resolution[1]);
    }

    if($width == 0 || $height == 0) throw new Exception('invalid total screen height or width');

    // switch modes
    $command = sprintf(
        "xrandr --output default --mode %sx%s --rate %s",
        $width, $height,
        $id// Nvidia "hack": use framerate as unique id
    );  
    echo "$command\n";
    passthru($command);

    sleep(5);

    // delete previous metamode
    delete_all_metamodes_except($id);

}

//exceptions = array ids not to delete
function delete_all_metamodes_except($exceptions = null)
{
    global $nvidia_metamodes;
    if(!is_array($exceptions)) $exceptions = array($exceptions);

    foreach($nvidia_metamodes as $id => $info) {
        if(!in_array($id, $exceptions)) {
            echo "Deleting metamode: {$info['metamode']}\n";
            $command = sprintf(
                "nv-control-dpy --delete-metamode '%s'",
                $info['metamode'] 
            );
            echo "$command\n";
            passthru($command);
        }
    }

    $nvidia_metamodes = get_metamodes();
}



// enable a display, and disable all others
// format: array(display => resolution, ... )
function enable_displays($enabled_displays){
    global $nvidia_displays, $nvidia_metamodes;
    $enabled_displays_upper = array();
    $displays = array();

    //make sure display names are uppercase and resolutions use lowercase 'x'
    foreach($enabled_displays as $name => $resolution) {
        $enabled_displays_upper[strtoupper($name)] = strtolower(trim($resolution));
        if(!in_array(strtoupper($name), array_keys($nvidia_displays))) {
            die("ERROR: Display '" . strtoupper($name) . "' is not connected! Aborting.\n");
        }
    }
    $enabled_displays = $enabled_displays_upper;

    // Now, we construct an array of all enabled and disabled displays, and get a metamode for it 

    // disable any displays not specified
    foreach(array_keys($nvidia_displays) as $cur_display) {
        if(array_key_exists(strtoupper($cur_display), $enabled_displays)) continue;

        $displays[] = array(
            'display' => $cur_display,
            'enabled' => false
        );
    }

    // enable specified displays
    foreach($enabled_displays as $name => $resolution) {
        $displays[] = array(
            'display' => strtoupper($name),
            'enabled' => true,
            'resolution' => $resolution
        );
    }

    $metamode_string = create_metamode_string($displays);

    $id = get_metamode($metamode_string);
    switch_mode($id);

    return $id;
}

// enable a display, and disable all others
function enable_single_display($display, $resolution = null){
    global $nvidia_displays, $nvidia_metamodes;

    $display = strtoupper($display);
    $resolution = strtolower($resolution);

    if(!in_array($display, array_keys($nvidia_displays))) {
        die("ERROR: Display '$display' is not connected! Aborting.\n");
    }

    $displays = array();

    // build "display" array to send to create_metamode_string
    foreach(array_keys($nvidia_displays) as $cur_display) {
        if($display == $cur_display) continue;

        $displays[] = array(
            'display' => $cur_display,
            'enabled' => false
        );
    }

    $displays[] = array(
        'display' => $display,
        'enabled' => true,
        'resolution' => $resolution
    );

    $metamode_string = create_metamode_string($displays);
    $id = get_metamode($metamode_string);
    switch_mode($id);

    return $id;
}

// quick and easy way to test if metamode strings are equivalent: hash, and then test for equality
// slow and crappy, but simple to implement using var_dump
function hash_metamode($metamode_string)
{
    $mode_array = parse_metamode(strtolower($metamode_string));
    if(empty($mode_array))
        throw new Exception('failed to parse metamode');
    if(!ksort($mode_array))
        throw new Exception('failed to sort metamode array');

    ob_start();
    var_dump($mode_array);
    $target_mode = ob_get_contents();
    ob_end_clean();

    return $target_mode;
}

function get_metamode($metamode_string) {
    global $nvidia_metamodes;

    $id = null;

    $target_mode = hash_metamode($metamode_string);

    foreach($nvidia_metamodes as $cur_id=>$mode_info) {
        $cur_mode = hash_metamode($mode_info['metamode']);

        if($target_mode == $cur_mode) {
            $id = $cur_id;
            break;
        }
    }

    if(!$id) {
        $id = add_metamode($metamode_string);
        $nvidia_metamodes = get_metamodes();
    }

    return $id;
}

function add_metamode($mode_string) {
    $command = "nv-control-dpy --add-metamode '$mode_string'";
    echo "$command\n";
    exec($command, $output);
    $output = implode("\r", $output);
    echo "$output\n";
    $id = preg_replace('/.*"id=(\d*)".*/', '${1}', $output);

    return $id;
}

function create_metamode_string($displays)
{
    global $nvidia_displays;

    $metamode = array();

    $associated = 0;
    foreach($displays as $display) {
        $associated |= $nvidia_displays[$display['display']];
        if($display['enabled'] || $display['resolution'] || $display['mode']) {
            $mode = ($display['mode'] ? $display['mode'] : 
                        ($display['resolution'] ? $display['resolution'] : 'nvidia-auto-select'));
            $resolution = ($display['resolution'] ? $display['resolution'] : 
                        ($display['mode'] ? '@'  . $display['mode'] : NV_DEFAULT_RESOLUTION)
                    );
            $offset = ($display['offset'] ? $display['offset'] : "+0+0");

            $metamode[] = sprintf(
                '%s: %s @%s %s',
                $display['display'],
                $mode,
                $resolution,
                $offset
            );
        
        } else {
            $metamode[] = sprintf(
                '%s: NULL',
                $display['display']
            );
        }
    }
    
    // set associated screens
    $command = sprintf(
        "nv-control-dpy --set-associated-dpys 0x%x",
        $associated
    );
    echo "$command\n";
    passthru($command);

    return implode(', ', $metamode);
}
?>
